@@ -23,10 +23,12 @@ try:
 except ImportError:
     import simplejson as json
 
-from riak.transports import RiakHttpTransport
 from riak.bucket import RiakBucket
 from riak.mapreduce import RiakMapReduce
 from riak.search import RiakSearch
+from riak.transports import RiakHttpTransport
+from riak.util import deprecated
+
 
 class RiakClient(object):
     """
@@ -36,7 +38,8 @@ class RiakClient(object):
     """
     def __init__(self, host='127.0.0.1', port=8098, prefix='riak',
                  mapred_prefix='mapred', transport_class=None,
-                 client_id=None, solr_transport_class=None):
+                 client_id=None, solr_transport_class=None,
+                 transport_options=None):
         """
         Construct a new ``RiakClient`` object.
 
@@ -52,24 +55,42 @@ class RiakClient(object):
         :type transport_class: :class:`RiakTransport`
         :param solr_transport_class: HTTP-based transport class for Solr interface queries
         :type transport_class: :class:`RiakHttpTransport`
+        :param transport_options: Optional key-value args to pass to the transport constuctor
+        :type transport_options: dict
         """
-        if not transport_class:
-            self._transport = RiakHttpTransport(host,
-                                                port,
-                                                prefix,
-                                                mapred_prefix,
-                                                client_id)
+        if transport_class is None:
+            transport_class = RiakHttpTransport
+
+        api = getattr(transport_class, 'api', 1)
+        if api >= 2:
+            hostports = [ (host, port), ]
+            self._cm = transport_class.default_cm(hostports)
+
+            # If no transport options are provided, then default to the
+            # empty dict, otherwise just pass through what we are provided.
+            if transport_options is None:
+                transport_options = {}
+
+            self._transport = transport_class(self._cm,
+                                              prefix=prefix,
+                                              mapred_prefix=mapred_prefix,
+                                              client_id=client_id,
+                                              **transport_options)
         else:
+            deprecated('please upgrade the transport to the new API')
+            self._cm = None
             self._transport = transport_class(host, port, client_id=client_id)
 
         self._r = "default"
         self._w = "default"
         self._dw = "default"
         self._rw = "default"
-        self._encoders = {'application/json':json.dumps,
-                          'text/json':json.dumps}
-        self._decoders = {'application/json':json.loads,
-                          'text/json':json.loads}
+        self._pr = "default"
+        self._pw = "default"
+        self._encoders = {'application/json': json.dumps,
+                          'text/json': json.dumps}
+        self._decoders = {'application/json': json.loads,
+                          'text/json': json.loads}
         self._solr = None
         self._host = host
         self._port = port
@@ -163,6 +184,46 @@ class RiakClient(object):
         self._rw = rw
         return self
 
+    def get_pr(self):
+        """
+        Get the PR-value setting for this ``RiakClient``. (default 0)
+
+        :rtype: integer
+        """
+        return self._pr
+
+    def set_pr(self, pr):
+        """
+        Set the PR-value for this ``RiakClient`` instance. See :func:`set_r` for a
+        description of how these values are used.
+
+        :param pr: The PR value.
+        :type pr: integer
+        :rtype: self
+        """
+        self._pr = pr
+        return self
+
+    def get_pw(self):
+        """
+        Get the PW-value setting for this ``RiakClient``. (default 0)
+
+        :rtype: integer
+        """
+        return self._pr
+
+    def set_pw(self, pw):
+        """
+        Set the PW-value for this ``RiakClient`` instance. See :func:`set_r` for a
+        description of how these values are used.
+
+        :param pw: The W value.
+        :type pw: integer
+        :rtype: self
+        """
+        self._pr = pr
+        return self
+
     def get_client_id(self):
         """
         Get the ``client_id`` for this ``RiakClient`` instance.
@@ -265,6 +326,16 @@ class RiakClient(object):
         mr = RiakMapReduce(self)
         return apply(mr.search, args)
 
+    def index(self, *args):
+        """
+        Start assembling a Map/Reduce operation based on secondary
+        index query results.
+
+        :rtype: :class:`RiakMapReduce`
+        """
+        mr = RiakMapReduce(self)
+        return apply(mr.index, args)
+
     def link(self, *args):
         """
         Start assembling a Map/Reduce operation. A shortcut for :func:`RiakMapReduce.link`.
