@@ -17,18 +17,39 @@ KIND, either express or implied.  See the License for the
 specific language governing permissions and limitations
 under the License.
 """
-import socket, struct
+from __future__ import with_statement
+
+import errno
+import socket
+import struct
 
 try:
     import json
 except ImportError:
     import simplejson as json
 
-from transport import RiakTransport
-from riak.metadata import *
-from riak.mapreduce import RiakMapReduce, RiakLink
 from riak import RiakError
-import riakclient_pb2
+from riak.mapreduce import RiakLink
+from riak.metadata import (
+        MD_CHARSET,
+        MD_CTYPE,
+        MD_ENCODING,
+        MD_INDEX,
+        MD_LASTMOD,
+        MD_LASTMOD_USECS,
+        MD_LINKS,
+        MD_USERMETA,
+        MD_VTAG,
+        )
+from riak.riak_index_entry import RiakIndexEntry
+from riak.transports import connection
+from riak.transports.transport import RiakTransport
+import riak.util
+
+try:
+    import riakclient_pb2
+except ImportError:
+    riakclient_pb2 = None
 
 ## Protocol codes
 MSG_CODE_ERROR_RESP           =  0
@@ -62,31 +83,91 @@ RIAKC_RW_QUORUM = 4294967293
 RIAKC_RW_ALL = 4294967292
 RIAKC_RW_DEFAULT = 4294967291
 
+# These are a specific set of socket errors
+# that could be raised on send/recv that indicate
+# that the socket is closed or reset, and is not
+# usable. On seeing any of these errors, the socket
+# should be closed, and the connection re-established.
+CONN_CLOSED_ERRORS = (
+                        errno.EHOSTUNREACH,
+                        errno.ECONNRESET,
+                        errno.EBADF,
+                        errno.EPIPE
+                     )
 
 
+class SocketWithId(connection.Socket):
+    def __init__(self, host, port):
+        super(SocketWithId, self).__init__(host, port)
+        self.last_client_id = None
+
+    def maybe_connect(self):
+        # If we're going to establish a new connection, then reset the last
+        # client_id used on this connection.
+        if self.sock is None:
+            self.last_client_id = None
+            super(SocketWithId, self).maybe_connect()
+
+    def send(self, pkt):
+        try:
+            self.sock.sendall(pkt)
+        except socket.error, e:
+            # If the socket is in a bad state, close it and allow it
+            # to re-connect on the next try
+            if e[0] in CONN_CLOSED_ERRORS:
+                self.close()
+            raise
+
+    def recv(self, want_len):
+        try:
+            res = self.sock.recv(want_len)
+
+            # Assume the socket is closed if no data is
+            # returned on a blocking read.
+            if len(res) == 0 and want_len > 0:
+                self.close()
+
+            return res
+
+        except socket.error, e:
+            # If the socket is in a bad state, close it and allow it
+            # to re-connect on the next try
+            if e[0] in CONN_CLOSED_ERRORS:
+                self.close()
+            raise
+
 
 class RiakPbcTransport(RiakTransport):
     """
     The RiakPbcTransport object holds a connection to the protocol buffers interface
     on the riak server.
     """
+
+    # We're using the new RiakTransport API
+    api = 2
+
     rw_names = {
-        'default' : RIAKC_RW_DEFAULT,
-        'all' : RIAKC_RW_ALL,
-        'quorum' : RIAKC_RW_QUORUM,
-        'one' : RIAKC_RW_ONE
+        'default': RIAKC_RW_DEFAULT,
+        'all': RIAKC_RW_ALL,
+        'quorum': RIAKC_RW_QUORUM,
+        'one': RIAKC_RW_ONE
         }
-    def __init__(self, host='127.0.0.1', port=8087, client_id=None):
+
+    # The ConnectionManager class that this transport prefers.
+    default_cm = connection.cm_using(SocketWithId)
+
+    def __init__(self, cm, client_id=None, max_attempts=1, **unused_options):
         """
         Construct a new RiakPbcTransport object.
-        @param string host - Hostname or IP address (default '127.0.0.1')
-        @param int port - Port number (default 8087)
         """
+        if riakclient_pb2 is None:
+            raise RiakError("this transport is not available (no protobuf)")
+
         super(RiakPbcTransport, self).__init__()
-        self._host = host
-        self._port = port
+
+        self._cm = cm
         self._client_id = client_id
-        self._sock = None
+        self._max_attempts = max_attempts
 
     def translate_rw_val(self, rw):
         val = self.rw_names.get(rw)
@@ -95,16 +176,15 @@ class RiakPbcTransport(RiakTransport):
         return val
 
     def __copy__(self):
-        return RiakPbcTransport(self._host, self._port)
+        return RiakPbcTransport(self._cm, self._client_id)
 
     def ping(self):
         """
         Ping the remote server
         @return boolean
         """
-        self.maybe_connect()
-        self.send_msg_code(MSG_CODE_PING_REQ)
-        msg_code, msg = self.recv_msg()
+        # An expected response code of None implies "any response is valid".
+        msg_code, msg = self.send_msg_code(MSG_CODE_PING_REQ, None)
         if msg_code == MSG_CODE_PING_RESP:
             return 1
         else:
@@ -114,13 +194,9 @@ class RiakPbcTransport(RiakTransport):
         """
         Get the client id used by this connection
         """
-        self.maybe_connect()
-        self.send_msg_code(MSG_CODE_GET_CLIENT_ID_REQ)
-        msg_code, resp = self.recv_msg()
-        if msg_code == MSG_CODE_GET_CLIENT_ID_RESP:
-            return resp.client_id
-        else:
-            raise RiakError("unexpected protocol buffer message code: %d"%msg_code)
+        msg_code, resp = self.send_msg_code(MSG_CODE_GET_CLIENT_ID_REQ,
+                                            MSG_CODE_GET_CLIENT_ID_RESP)
+        return resp.client_id
 
     def set_client_id(self, client_id):
         """
@@ -129,15 +205,23 @@ class RiakPbcTransport(RiakTransport):
         req = riakclient_pb2.RpbSetClientIdReq()
         req.client_id = client_id
 
-        self.maybe_connect()
-        self.send_msg(MSG_CODE_SET_CLIENT_ID_REQ, req)
-        msg_code, resp = self.recv_msg()
-        if msg_code == MSG_CODE_SET_CLIENT_ID_RESP:
-            return True
-        else:
-            raise RiakError("unexpected protocol buffer message code: %d"%msg_code)
+        msg_code, resp = self.send_msg(MSG_CODE_SET_CLIENT_ID_REQ, req,
+                                       MSG_CODE_SET_CLIENT_ID_RESP)
 
-    def get(self, robj, r = None, vtag = None):
+        # Using different client_id values across connections is a bad idea
+        # since you never know which connection you might use for a given
+        # API call. Setting the client_id manually (rather than as part of
+        # the transport construction) can be error-prone since the connection
+        # could drop and be reinstated using self._client_id.
+        #
+        # To minimize the potential impact of variant client_id values across
+        # connections, we'll store this new client_id and use it for all
+        # future connections.
+        self._client_id = client_id
+
+        return True
+
+    def get(self, robj, r=None, pr=None, vtag=None):
         """
         Serialize get request and deserialize response
         """
@@ -148,22 +232,22 @@ class RiakPbcTransport(RiakTransport):
 
         req = riakclient_pb2.RpbGetReq()
         req.r = self.translate_rw_val(r)
+        req.pr = self.translate_rw_val(pr)
 
         req.bucket = bucket.get_name()
         req.key = robj.get_key()
 
-        self.maybe_connect()
-        self.send_msg(MSG_CODE_GET_REQ, req)
-        msg_code, resp = self.recv_msg()
+        # An expected response code of None implies "any response is valid".
+        msg_code, resp = self.send_msg(MSG_CODE_GET_REQ, req, None)
         if msg_code == MSG_CODE_GET_RESP:
             contents = []
             for c in resp.content:
                 contents.append(self.decode_content(c))
             return resp.vclock, contents
         else:
-            return 0
+            return None
 
-    def put(self, robj, w = None, dw = None, return_body = True):
+    def put(self, robj, w=None, dw=None, pw=None, return_body=True, if_none_match=False):
         """
         Serialize get request and deserialize response
         """
@@ -172,8 +256,12 @@ class RiakPbcTransport(RiakTransport):
         req = riakclient_pb2.RpbPutReq()
         req.w = self.translate_rw_val(w)
         req.dw = self.translate_rw_val(dw)
+        req.pw = self.translate_rw_val(pw)
+
         if return_body:
             req.return_body = 1
+        if if_none_match:
+            req.if_none_match = 1
 
         req.bucket = bucket.get_name()
         req.key = robj.get_key()
@@ -183,18 +271,49 @@ class RiakPbcTransport(RiakTransport):
 
         self.pbify_content(robj.get_metadata(), robj.get_encoded_data(), req.content)
 
-        self.maybe_connect()
-        self.send_msg(MSG_CODE_PUT_REQ, req)
-        msg_code, resp = self.recv_msg()
-        if msg_code != MSG_CODE_PUT_RESP:
-            raise RiakError("unexpected protocol buffer message code: %d"%msg_code)
+        msg_code, resp = self.send_msg(MSG_CODE_PUT_REQ, req,
+                                       MSG_CODE_PUT_RESP)
         if resp is not None:
             contents = []
             for c in resp.content:
                 contents.append(self.decode_content(c))
             return resp.vclock, contents
 
-    def delete(self, robj, rw = None):
+    def put_new(self, robj, w=None, dw=None, pw=None, return_body=True, if_none_match=False):
+        """Put a new object into the Riak store, returning its (new) key.
+
+        If return_meta is False, then the vlock and metadata return values
+        will be None.
+
+        @return (key, vclock, metadata)
+        """
+        bucket = robj.get_bucket()
+
+        req = riakclient_pb2.RpbPutReq()
+        req.w = self.translate_rw_val(w)
+        req.dw = self.translate_rw_val(dw)
+        req.pw = self.translate_rw_val(pw)
+
+        if return_body:
+            req.return_body = 1
+        if if_none_match:
+            req.if_none_match = 1
+
+        req.bucket = bucket.get_name()
+
+        self.pbify_content(robj.get_metadata(), robj.get_encoded_data(), req.content)
+
+        msg_code, resp = self.send_msg(MSG_CODE_PUT_REQ, req,
+                                       MSG_CODE_PUT_RESP)
+        if not resp:
+            raise RiakError("missing response object")
+        if len(resp.content) != 1:
+            raise RiakError("siblings were returned from object creation")
+
+        metadata, content = self.decode_content(resp.content[0])
+        return resp.key, resp.vclock, metadata
+
+    def delete(self, robj, rw=None, r=None, w=None, dw=None, pr=None, pw=None):
         """
         Serialize get request and deserialize response
         """
@@ -202,15 +321,19 @@ class RiakPbcTransport(RiakTransport):
 
         req = riakclient_pb2.RpbDelReq()
         req.rw = self.translate_rw_val(rw)
+        req.r = self.translate_rw_val(r)
+        req.w = self.translate_rw_val(w)
+        req.dw = self.translate_rw_val(dw)
+        req.pr = self.translate_rw_val(pr)
+        req.pw = self.translate_rw_val(pw)
+
+        # TODO: Set the vclock if present
 
         req.bucket = bucket.get_name()
         req.key = robj.get_key()
 
-        self.maybe_connect()
-        self.send_msg(MSG_CODE_DEL_REQ, req)
-        msg_code, resp = self.recv_msg()
-        if msg_code != MSG_CODE_DEL_RESP:
-            raise RiakError("unexpected protocol buffer message code: %d"%msg_code)
+        msg_code, resp = self.send_msg(MSG_CODE_DEL_REQ, req,
+                                       MSG_CODE_DEL_RESP)
         return self
 
     def get_keys(self, bucket):
@@ -220,19 +343,12 @@ class RiakPbcTransport(RiakTransport):
         req = riakclient_pb2.RpbListKeysReq()
         req.bucket = bucket.get_name()
 
-        self.maybe_connect()
-        self.send_msg(MSG_CODE_LIST_KEYS_REQ, req)
         keys = []
-        while True:
-            msg_code, resp = self.recv_msg()
-            if msg_code != MSG_CODE_LIST_KEYS_RESP:
-                raise RiakError("unexpected protocol buffer message code: %d"%msg_code)
-
+        def _handle_response(resp):
             for key in resp.keys:
                 keys.append(key)
-
-            if resp.HasField("done") and resp.done:
-                break
+        self.send_msg_multi(MSG_CODE_LIST_KEYS_REQ, req,
+                            MSG_CODE_LIST_KEYS_RESP, _handle_response)
 
         return keys
 
@@ -240,11 +356,8 @@ class RiakPbcTransport(RiakTransport):
         """
         Serialize bucket listing request and deserialize response
         """
-        self.maybe_connect()
-        self.send_msg_code(MSG_CODE_LIST_BUCKETS_REQ)
-        msg_code, resp = self.recv_msg()
-        if msg_code != MSG_CODE_LIST_BUCKETS_RESP:
-          raise RiakError("unexpected protocol buffer message code: %d"%msg_code)
+        msg_code, resp = self.send_msg_code(MSG_CODE_LIST_BUCKETS_REQ,
+                                            MSG_CODE_LIST_BUCKETS_RESP)
         return resp.buckets
 
     def get_bucket_props(self, bucket):
@@ -254,11 +367,8 @@ class RiakPbcTransport(RiakTransport):
         req = riakclient_pb2.RpbGetBucketReq()
         req.bucket = bucket.get_name()
 
-        self.maybe_connect()
-        self.send_msg(MSG_CODE_GET_BUCKET_REQ, req)
-        msg_code, resp = self.recv_msg()
-        if msg_code != MSG_CODE_GET_BUCKET_RESP:
-            raise RiakError("unexpected protocol buffer message code: %d"%msg_code)
+        msg_code, resp = self.send_msg(MSG_CODE_GET_BUCKET_REQ, req,
+                                       MSG_CODE_GET_BUCKET_RESP)
         props = {}
         if resp.props.HasField('n_val'):
             props['n_val'] = resp.props.n_val
@@ -267,31 +377,27 @@ class RiakPbcTransport(RiakTransport):
 
         return props
 
-
     def set_bucket_props(self, bucket, props):
         """
         Serialize set bucket property request and deserialize response
         """
         req = riakclient_pb2.RpbSetBucketReq()
         req.bucket = bucket.get_name()
-        if not 'n_val' in props and not 'allow_mult' in props: return self
+        if not 'n_val' in props and not 'allow_mult' in props:
+            return self
 
         if 'n_val' in props:
             req.props.n_val = props['n_val']
         if 'allow_mult' in props:
             req.props.allow_mult = props['allow_mult']
 
-        self.maybe_connect()
-        self.send_msg(MSG_CODE_SET_BUCKET_REQ, req)
-        msg_code, resp = self.recv_msg()
-        if msg_code != MSG_CODE_SET_BUCKET_RESP:
-            raise RiakError("unexpected protocol buffer message code: %d"%msg_code)
-
+        msg_code, resp = self.send_msg(MSG_CODE_SET_BUCKET_REQ, req,
+                                       MSG_CODE_SET_BUCKET_RESP)
         return self
 
     def mapred(self, inputs, query, timeout=None):
         # Construct the job, optionally set the timeout...
-        job = {'inputs':inputs, 'query':query}
+        job = {'inputs': inputs, 'query': query}
         if timeout is not None:
             job['timeout'] = timeout
 
@@ -301,25 +407,18 @@ class RiakPbcTransport(RiakTransport):
         req.request = content
         req.content_type = "application/json"
 
-        self.maybe_connect()
-        self.send_msg(MSG_CODE_MAPRED_REQ, req)
-
         # dictionary of phase results - each content should be an encoded array
         # which is appended to the result for that phase.
         result = {}
-        while True:
-            msg_code, resp = self.recv_msg()
-            if msg_code != MSG_CODE_MAPRED_RESP:
-                raise RiakError("unexpected protocol buffer message code: %d"%msg_code)
+        def _handle_response(resp):
             if resp.HasField("phase") and resp.HasField("response"):
                 content = json.loads(resp.response)
                 if resp.phase in result:
                     result[resp.phase] += content
                 else:
                     result[resp.phase] = content
-
-            if resp.HasField("done") and resp.done:
-                break;
+        self.send_msg_multi(MSG_CODE_MAPRED_REQ, req, MSG_CODE_MAPRED_RESP,
+                            _handle_response)
 
         # If a single result - return the same as the HTTP interface does
         # otherwise return all the phase information
@@ -330,23 +429,10 @@ class RiakPbcTransport(RiakTransport):
         else:
             return result
 
-
-    def maybe_connect(self):
-        if self._sock is None:
-            self._sock = s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-
-            try:
-                s.connect((self._host, self._port))
-            except:
-                self._sock = None
-                raise
-
-            if self._client_id:
-                self.set_client_id(self._client_id)
-
-    def send_msg_code(self, msg_code):
-        pkt = struct.pack("!iB", 1, msg_code)
-        self._sock.send(pkt)
+    def send_msg_code(self, msg_code, expect):
+        with self._cm.withconn() as conn:
+            self.send_pkt(conn, struct.pack("!iB", 1, msg_code))
+            return self.recv_msg(conn, expect)
 
     def encode_msg(self, msg_code, msg):
         str = msg.SerializeToString()
@@ -354,15 +440,53 @@ class RiakPbcTransport(RiakTransport):
         hdr = struct.pack("!iB", 1 + slen, msg_code)
         return hdr + str
 
-    def send_msg(self, msg_code, msg):
-        pkt = self.encode_msg(msg_code, msg)
-        sent_len = self._sock.send(pkt)
-        if sent_len != len(pkt):
-            raise RiakError("PB socket returned short write {0} - expected {1}".
-                            format(sent_len, len(pkt)))
+    def send_msg(self, msg_code, msg, expect):
+        with self._cm.withconn() as conn:
+            self.send_pkt(conn, self.encode_msg(msg_code, msg))
+            if msg_code == MSG_CODE_SET_CLIENT_ID_REQ:
+                conn.last_client_id = self._client_id
+            return self.recv_msg(conn, expect)
 
-    def recv_msg(self):
-        self.recv_pkt()
+    def send_msg_multi(self, msg_code, msg, expect, handler):
+        with self._cm.withconn() as conn:
+            self.send_pkt(conn, self.encode_msg(msg_code, msg))
+            while True:
+                msg_code, resp = self.recv_msg(conn, expect)
+                handler(resp)
+                if resp.HasField("done") and resp.done:
+                    break
+
+    def send_pkt(self, conn, pkt):
+        attempt = 0
+        e = None
+        for attempt in xrange(self._max_attempts):
+            e = None
+            try:
+                conn.maybe_connect()
+
+                # If the last client_id used on this connection is different than our
+                # client_id, then set a new ID on the connection.
+                if conn.last_client_id != self._client_id:
+                    req = riakclient_pb2.RpbSetClientIdReq()
+                    req.client_id = self._client_id
+                    conn.send(self.encode_msg(MSG_CODE_SET_CLIENT_ID_REQ, req))
+                    conn.last_client_id = self._client_id
+                    self.recv_msg(conn, MSG_CODE_SET_CLIENT_ID_RESP)
+
+                conn.send(pkt)
+                break
+            except socket.error, e:
+                # If this is some unknown socket error bail out
+                # instead of retrying
+                if e[0] not in CONN_CLOSED_ERRORS:
+                    raise
+
+        # Max attempts reached, raise whatever exception we are getting
+        if attempt + 1 == self._max_attempts and e is not None:
+            raise e
+
+    def recv_msg(self, conn, expect):
+        self.recv_pkt(conn)
         msg_code, = struct.unpack("B", self._inbuf[:1])
         if msg_code == MSG_CODE_ERROR_RESP:
             msg = riakclient_pb2.RpbErrorResp()
@@ -398,27 +522,29 @@ class RiakPbcTransport(RiakTransport):
             msg = riakclient_pb2.RpbMapRedResp()
             msg.ParseFromString(self._inbuf[1:])
         else:
-            raise Exception("unknown msg code {0}".format(msg_code))
+            raise Exception("unknown msg code %s" % msg_code)
+        if expect and msg_code != expect:
+            raise RiakError("unexpected protocol buffer message code: %d"
+                            % msg_code)
         return msg_code, msg
 
-
-    def recv_pkt(self):
-        nmsglen = self._sock.recv(4)
+    def recv_pkt(self, conn):
+        nmsglen = conn.recv(4)
         if len(nmsglen) != 4:
-            self._sock = None
-            raise RiakError("Socket returned short packet length {0} - expected 4".
-                            format(nmsglen))
+            raise RiakError("Socket returned short packet length %d - expected 4"
+                            % len(nmsglen))
         msglen, = struct.unpack('!i', nmsglen)
         self._inbuf_len = msglen
         self._inbuf = ''
         while len(self._inbuf) < msglen:
             want_len = min(8192, msglen - len(self._inbuf))
-            recv_buf = self._sock.recv(want_len)
-            if not recv_buf: break
+            recv_buf = conn.recv(want_len)
+            if not recv_buf:
+                break
             self._inbuf += recv_buf
         if len(self._inbuf) != self._inbuf_len:
-            raise RiakError("Socket returned short packet {0} - expected {1}".
-                            format(len(self._inbuf), self._inbuf_len))
+            raise RiakError("Socket returned short packet %d - expected %d"
+                            % (len(self._inbuf), self._inbuf_len))
 
     def decode_contents(self, rpb_contents):
         contents = []
@@ -462,12 +588,18 @@ class RiakPbcTransport(RiakTransport):
             usermeta[usermd.key] = usermd.value
         if len(usermeta) > 0:
             metadata[MD_USERMETA] = usermeta
+        indexes = []
+        for index in rpb_content.indexes:
+            rie = RiakIndexEntry(index.key, index.value)
+            indexes.append(rie)
+        if len(indexes) > 0:
+            metadata[MD_INDEX] = indexes
         return metadata, rpb_content.value
 
-    def pbify_content(self, metadata, data, rpb_content) :
+    def pbify_content(self, metadata, data, rpb_content):
         # Convert the broken out fields, building up
         # pbmetadata for any unknown ones
-        for k,v in metadata.iteritems():
+        for k, v in metadata.iteritems():
             if k == MD_CTYPE:
                 rpb_content.content_type = v
             elif k == MD_CHARSET:
@@ -479,6 +611,11 @@ class RiakPbcTransport(RiakTransport):
                     pair = rpb_content.usermeta.add()
                     pair.key = uk
                     pair.value = uv
+            elif k == MD_INDEX:
+                for rie in v:
+                    pair = rpb_content.indexes.add()
+                    pair.key = rie.get_field()
+                    pair.value = rie.get_value()
             elif k == MD_LINKS:
                 for link in v:
                     pb_link = rpb_content.links.add()
@@ -487,119 +624,3 @@ class RiakPbcTransport(RiakTransport):
                     pb_link.tag = link.get_tag()
         rpb_content.value = data
 
-from Queue import Empty, Full, Queue
-import contextlib
-class RiakPbcCachedTransport(RiakTransport):
-    """Threadsafe pool of PBC connections, based on urllib3's pool [aka Queue]"""
-    def __init__(self, host='127.0.0.1', port=8087, client_id=None, maxsize=0, block=False, timeout=None):
-        self.host = host
-        self.port = port
-        self.client_id = client_id
-        self.block = block
-        self.timeout = timeout
-
-        self.pool = Queue(maxsize)
-        # Fill the queue up so that doing get() on it will block properly (check Queue#get)
-        [self.pool.put(None) for _ in xrange(maxsize)]
-
-    def _new_connection(self):
-        """New PBC connection"""
-        return RiakPbcTransport(self.host, self.port, self.client_id)
-
-    def _get_connection(self):
-        connection = None
-        try:
-            connection = self.pool.get(block=self.block, timeout=self.timeout)
-        except Empty:
-            pass
-        return connection or self._new_connection()
-
-    def _put_connection(self, connection):
-        try:
-            self.pool.put(connection, block=False)
-        except Full:
-            pass
-
-    @contextlib.contextmanager
-    def _get_connection_from_pool(self):
-        """checkout conn, try operation, put conn back in pool"""
-        connection = self._get_connection()
-        try:
-            yield connection
-        finally:
-            self._put_connection(connection)
-
-    def ping(self):
-        """
-        Ping the remote server
-        @return boolean
-        """
-        with self._get_connection_from_pool() as connection:
-            return connection.ping()
-
-    def get(self, robj, r = None, vtag = None):
-        """
-        Serialize get request and deserialize response
-        @return (vclock=None, [(metadata, value)]=None)
-        """
-        with self._get_connection_from_pool() as connection:
-            return connection.get(robj, r, vtag)
-
-    def put(self, robj, w = None, dw = None, return_body = True):
-        """
-        Serialize put request and deserialize response - if 'content'
-        is true, retrieve the updated metadata/content
-        @return (vclock=None, [(metadata, value)]=None)
-        """
-        with self._get_connection_from_pool() as connection:
-            return connection.put(robj, w, dw, return_body)
-
-    def delete(self, robj, rw = None):
-        """
-        Serialize delete request and deserialize response
-        @return true
-        """
-        with self._get_connection_from_pool() as connection:
-            return connection.delete(robj, rw)
-
-    def get_buckets(self):
-        """
-        Serialize bucket listing request and deserialize response
-        """
-        with self._get_connection_from_pool() as connection:
-            return connection.get_buckets()
-
-    def get_bucket_props(self, bucket) :
-        """
-        Serialize get bucket property request and deserialize response
-        @return dict()
-        """
-        with self._get_connection_from_pool() as connection:
-            return connection.get_bucket_props(bucket)
-
-    def set_bucket_props(self, bucket, props) :
-        """
-        Serialize set bucket property request and deserialize response
-        bucket = bucket object
-        props = dictionary of properties
-        @return boolean
-        """
-        with self._get_connection_from_pool() as connection:
-            return connection.set_bucket_props(bucket, props)
-
-    def mapred(self, inputs, query, timeout = None) :
-        """
-        Serialize map/reduce request
-        """
-        with self._get_connection_from_pool() as connection:
-            return connection.mapred(inputs, query, timeout)
-
-    def set_client_id(self, client_id):
-        """Mmm, this can turn ugly if you use different id for different objects in the pool"""
-        with self._get_connection_from_pool() as connection:
-            return connection.set_client_id(client_id)
-
-    def get_client_id(self):
-        """see set_client_id notes, you can do wrong with this"""
-        with self._get_connection_from_pool() as connection:
-            return connection.get_client_id()
