@@ -29,13 +29,19 @@ class RiakTransport(object):
     Class to encapsulate transport details
     """
 
+    # Subclasses should specify their API level.
+    #   * missing or 1: the API used up and through 1.3.x.
+    #   * 2: the API introduced with 1.4.x
+    #
+    # api = 2
+
     @classmethod
     def make_random_client_id(self):
         """
         Returns a random client identifier
         """
         return 'py_%s' % base64.b64encode(
-                str(random.randint(1, 1073741824)))
+                str(random.randint(1, 0x40000000)))
 
     @classmethod
     def make_fixed_client_id(self):
@@ -69,6 +75,16 @@ class RiakTransport(object):
         """
         raise RiakError("not implemented")
 
+    def put_new(self, robj, w=None, dw=None, return_meta=True):
+        """Put a new object into the Riak store, returning its (new) key.
+
+        If return_meta is False, then the vlock and metadata return values
+        will be None.
+
+        @return (key, vclock, metadata)
+        """
+        raise RiakError("not implemented")
+
     def delete(self, robj, rw = None):
         """
         Serialize delete request and deserialize response
@@ -134,9 +150,9 @@ class RiakTransport(object):
         """
         raise RiakError("luwak not supported by this transport.")
 
+    def delete_file(self, key):
         """
         Delete an object in luwak.
         key = the object's key
         """
-    def delete_file(self, key):
         raise RiakError("luwak not supported by this transport.")
